/**************************************************************
 This file is part of
    Morphological Curvature Flow Approximation for AviSynth
                       by Wolfgang Boiger
 Copyright (C) 2007 Wolfgang Boiger, Berlin
 
 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 version 2, as published by the Free Software Foundation in
 June 1991 (see 'GNUGPLv2.txt').
 
 There is no guaranty that the rights of third parties
 (including software patents) are not infringed upon by the
 execution or distribution of this program.
 Referring to section 8 of the General Public License, any use
 of this program (including all activities mentioned in
 section 0 and including execution) is restricted to countries
 where any of these activities do not infringe upon any such
 rights of third parties. It lies with the user to verify the
 compliance of his or her use (especially concerning but not
 limited to distribution) of this program with the rights of
 any third parties in the respective countries or regions.
 
 This program is distributed in the hope that it will be
 useful, but WITHOUT ANY WARRANTY; without even the implied
 warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 PURPOSE.  See the GNU General Public License for more details.
 
 You should have received a copy of the GNU General Public
 License 2 along with this program; if not, write to the
   Free Software Foundation, Inc.,
   51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 **************************************************************/



/**************************************************************
 This file holds all functions which have to be implemented
 multiple times since they depend on the colorspace (and some
 other things).
 **************************************************************/



/**************************************************************
                           PickPixel
 AVSCurveFlow_CSdep_PickPixel
 AVSCurveFlow_CSdep_Stride
 AVSCurveFlow_CSdep_Table2
 AVSCurveFlow_CSdep_WidthEven
 **************************************************************/

#ifdef  AVSCurveFlow_CSdep_PickPixel
void  AVSCurveFlow_CSdep_PickPixel(
  unsigned char  PickData[],
  const  BYTE  SrcLuma[],
  int  Width,
  int  Height,
  int  SrcPitch,
  const  unsigned char  Table1[]
  #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
  ,  const  unsigned char  Table2[]
  #endif
  )
{
  register  int x, y;
  signed int  xm, xpp, ym, yp;
  BYTE  Pixel1[9];
  BYTE  Pixel2[9];
  long int  Index1, Index2;
  #if (defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0))
  #define  xp  AVSCurveFlow_CSdep_Stride
  #else
  signed int  xp;
  #endif


  ym = 0;  /* ym and yp store offsets for previous/next pixel line (0 at borders => image appears mirrored) */
  yp = SrcPitch;
  for (y=0 ; y<Height ; y++)
  {
    if (y==1)  ym = -SrcPitch;
    if (y==Height-1)  yp = 0;
    /* This part *must* be changed when frames with odd width are allowed! */
    xm = 0;
    #if (!(defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0)))
    xp = AVSCurveFlow_CSdep_Stride;
    #endif
    xpp = 2*AVSCurveFlow_CSdep_Stride;
    for (x=0 ; x<Width ; x+=2)
    {
      if (x==2)  xm = -AVSCurveFlow_CSdep_Stride;
      if (x==Width-2)  xpp = +AVSCurveFlow_CSdep_Stride;
      #if (!(defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0)))
       else  if (x==Width-1)  xp = xpp = 0;
      #endif
      /******************************************************************
        Location of pixels in the 3x4 Pixel Block:
             P1[6]      P1[0]=P2[0]   P1[3]=P2[3]      P2[6]
             P1[7]      P1[1]=P2[1]   P1[4]=P2[4]      P2[7]
             P1[8]      P1[2]=P2[2]   P1[5]=P2[5]      P2[8]
        Which bits of the table index are important for which block?:
           \------------most sign. bits-----------/
                      \---------least sign. bits-------------/
       ******************************************************************/
      Pixel1[0] = SrcLuma[ym    ];
      Pixel1[1] = SrcLuma[   0  ];
      Pixel1[2] = SrcLuma[yp    ];
      Pixel1[3] = SrcLuma[ym+xp ];
      Pixel1[4] = SrcLuma[   xp ];
      Pixel1[5] = SrcLuma[yp+xp ];
      Pixel1[6] = SrcLuma[ym+xm ];
      Pixel1[7] = SrcLuma[   xm ];
      Pixel1[8] = SrcLuma[yp+xm ];
      Index1 = InsertSort(Pixel1,0,1,6);
      #if (!(defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0)))
      if (x!=Width-1)
      #endif
      {
        memcpy(Pixel2,Pixel1,sizeof(BYTE)*6);
        Pixel2[6] = SrcLuma[ym+xpp];
        Pixel2[7] = SrcLuma[   xpp];
        Pixel2[8] = SrcLuma[yp+xpp];
        Index2 = InsertSort(Pixel2,Index1,6,9);
        #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
        PickData[1] = Table1[Index2]%16+16*(Table2[Index2]%16);
        #else
        PickData[1] = Table1[Index2]%16;
        #endif
        SrcLuma += AVSCurveFlow_CSdep_Stride;
      }
      SrcLuma += AVSCurveFlow_CSdep_Stride;
      Index1 = InsertSort(Pixel1,Index1,6,9);
      #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
      PickData[0] = Table1[Index1]/16+16*(Table2[Index1]/16);
      #else
      PickData[0] = Table1[Index1]/16;
      #endif
      #if (!(defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0)))
      PickData += (x!=Width-1)?2:1;
      #else
      PickData += 2;
      #endif
    }
    SrcLuma += SrcPitch-AVSCurveFlow_CSdep_Stride*Width;
  }

  #if (defined(AVSCurveFlow_CSdep_WidthEven) && (AVSCurveFlow_CSdep_WidthEven!=0))
  #undef  xp
  #endif
}
#undef  AVSCurveFlow_CSdep_PickPixel
#undef  AVSCurveFlow_CSdep_Stride
#undef  AVSCurveFlow_CSdep_Table2
#undef  AVSCurveFlow_CSdep_WidthEven
#endif



/**************************************************************
                          ChangePixel
 AVSCurveFlow_CSdep_ChangePixel
 AVSCurveFlow_CSdep_Stride
 AVSCurveFlow_CSdep_ChromaWidth
 AVSCurveFlow_CSdep_ChromaHeight
 AVSCurveFlow_CSdep_Table2
 **************************************************************/

#ifdef  AVSCurveFlow_CSdep_ChangePixel
void  AVSCurveFlow_CSdep_ChangePixel(BYTE  DestData[],  const  BYTE  SrcData[],  const  unsigned char  PickData[],  int  Width,  int  Height,  int  DestPitch,  int  SrcPitch)
{
  /* If one chroma pixel spans multiple luma rows, n counts those rows */
  #if (AVSCurveFlow_CSdep_ChromaHeight==1)
  #define n 0
  #else
  int  n;
  #endif
  /* Likewise, m counts the columns spaned by one chroma pixel */
  #if (AVSCurveFlow_CSdep_ChromaWidth==1)
  #define m 0
  #else
  int  m;
  #endif
  register  int  x, y;
  signed int  xc[3], yc[3];  /* Offsets of current position to address neighbor pixels */
  long  Value;  /* temporary variable to store current value of (big chroma) pixel */
  unsigned char  P;  /* component of PickData */


  xc[1] = yc[1] = 0;
  #if (AVSCurveFlow_CSdep_ChromaHeight==1)
  yc[0] = 0;
  yc[2] = SrcPitch;
  #endif
  for (y=0 ; y<Height/AVSCurveFlow_CSdep_ChromaHeight ; y++)
  {
    #if (AVSCurveFlow_CSdep_ChromaHeight==1)
    if (y==1)  yc[0] = -SrcPitch;
    if (y==Height-1)  yc[2] = 0;
    #endif
    #if (AVSCurveFlow_CSdep_ChromaWidth==1)
    xc[0] = 0;
    xc[2] = AVSCurveFlow_CSdep_Stride;
    #endif
    for (x=0 ; x<Width/AVSCurveFlow_CSdep_ChromaWidth ; x++)
    {
      #if (AVSCurveFlow_CSdep_ChromaWidth==1)
      if (x==1)  xc[0] = -AVSCurveFlow_CSdep_Stride;
      if (x==Width-1)  xc[2] = 0;
      #endif
      #if ((AVSCurveFlow_CSdep_ChromaHeight>2)&&(AVSCurveFlow_CSdep_ChromaWidth>2))
      /* If results of two look-up tables have to be merged, it is done by adding up all values and dividing by 2 later */
      #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
      Value = 2*(AVSCurveFlow_CSdep_ChromaHeight-2)*(AVSCurveFlow_CSdep_ChromaWidth-2)*SrcChroma[AVSCurveFlow_CSdep_ChromaStride*x];
      #else
      Value = (AVSCurveFlow_CSdep_ChromaHeight-2)*(AVSCurveFlow_CSdep_ChromaWidth-2)*SrcChroma[AVSCurveFlow_CSdep_ChromaStride*x];
      #endif
      #elif (!((AVSCurveFlow_CSdep_ChromaHeight==1)&&(AVSCurveFlow_CSdep_ChromaWidth==1)))
      Value = 0;
      #endif
      #if (AVSCurveFlow_CSdep_ChromaHeight!=1)
      for (n=0 ; n<AVSCurveFlow_CSdep_ChromaHeight ; n++)
      {
        yc[0] = ((n==0) && (y>0))?(-SrcPitch):0;
        yc[2] = ((n==AVSCurveFlow_CSdep_ChromaHeight-1) && (y<Height/AVSCurveFlow_CSdep_ChromaHeight-1))?(+SrcPitch):0;
      #else
      {
      #endif
        #if (AVSCurveFlow_CSdep_ChromaWidth!=1)
        for (m=0 ; m<AVSCurveFlow_CSdep_ChromaWidth ; m++)
        {
          xc[0] = ((m==0) && (x>0))?(-AVSCurveFlow_CSdep_Stride):0;
          xc[2] = ((m==AVSCurveFlow_CSdep_ChromaWidth-1) && (x<Width/AVSCurveFlow_CSdep_ChromaWidth-1))?(+AVSCurveFlow_CSdep_Stride):0;
        #else
        {
        #endif
          #if ((AVSCurveFlow_CSdep_ChromaHeight>2)&&(AVSCurveFlow_CSdep_ChromaWidth>2))
          if ((n!=0) && (m!=0) && (n!=AVSCurveFlow_CSdep_ChromaHeight-1) && (m!=AVSCurveFlow_CSdep_ChromaWidth-1))  continue;
          #endif
          /************************************************************************
            Content of P: Each 2 bits store a Row/Column offset, increased with 1.
                | Table2-Row | Table2-Column | Table1-Row | Table1-Column |
           ************************************************************************/
          P = PickData[m+Width*n]%16;
          #if ((AVSCurveFlow_CSdep_ChromaHeight==1)&&(AVSCurveFlow_CSdep_ChromaWidth==1))
          Value = SrcData[yc[P/4]+xc[P%4]];
          #else
          Value += SrcData[yc[P/4]+xc[P%4]];
          #endif
          #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
          P = PickData[m+Width*n]/16;
          Value += SrcData[yc[P/4]+xc[P%4]];
          #endif
        }
      }
      #if ((AVSCurveFlow_CSdep_ChromaHeight==1)&&(AVSCurveFlow_CSdep_ChromaWidth==1))
      #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
      DestData[0] = Value/2;   /*+(Value%4==3);  Unbiased rounding: Round to even number */
      #else
      DestData[0] = Value;
      #endif
      #else
      #if (defined(AVSCurveFlow_CSdep_Table2) && (AVSCurveFlow_CSdep_Table2!=0))
      DestData[0] = round(0.5*Value/AVSCurveFlow_CSdep_ChromaHeight/AVSCurveFlow_CSdep_ChromaWidth);
      #else
      DestData[0] = round(1.0*Value/AVSCurveFlow_CSdep_ChromaHeight/AVSCurveFlow_CSdep_ChromaWidth);
      #endif
      #endif
      DestData += AVSCurveFlow_CSdep_Stride;
      SrcData += AVSCurveFlow_CSdep_Stride;
      PickData += AVSCurveFlow_CSdep_ChromaWidth;
    }
    DestData += DestPitch-(Width*AVSCurveFlow_CSdep_Stride)/AVSCurveFlow_CSdep_ChromaWidth;
    SrcData += SrcPitch-(Width*AVSCurveFlow_CSdep_Stride)/AVSCurveFlow_CSdep_ChromaWidth;
  }
  #ifdef n
  #undef n
  #endif
  #ifdef m
  #undef m
  #endif
}
#undef  AVSCurveFlow_CSdep_ChangePixel
#undef  AVSCurveFlow_CSdep_Stride
#undef  AVSCurveFlow_CSdep_ChromaWidth
#undef  AVSCurveFlow_CSdep_ChromaHeight
#undef  AVSCurveFlow_CSdep_Table2
#endif
